// Ex 4 - Cartesian Paths

#include <moveit/move_group_interface/move_group_interface.h>
#include <moveit/planning_scene_interface/planning_scene_interface.h>

#include <moveit_msgs/DisplayRobotState.h>
#include <moveit_msgs/DisplayTrajectory.h>

#include "moveit_msgs/CollisionObject.h"
#include "moveit_msgs/MoveItErrorCodes.h" //Added to include the SUCCESS etc. error codes #include <moveit_msgs/AttachedCollisionObject.h>

#include <moveit_visual_tools/moveit_visual_tools.h>
#include <rviz_visual_tools/rviz_visual_tools.h>

#include <ros/duration.h>
#include <ros/ros.h>
#include <ros/time.h>
#include <std_msgs/Float64.h>
#include <std_msgs/Float64MultiArray.h>

int main(int argc, char **argv) {
  // ros::init(argc, argv, "move_group_interface_tutorial");
  ros::init(argc, argv, "r7_test");
  ros::NodeHandle node_handle;
  ros::AsyncSpinner spinner(1);
  spinner.start();

  // BEGIN_TUTORIAL
  //
  // Setup
  // ^^^^^
  // MoveIt! operates on sets of joints called "planning groups" and stores them
  // in an object called the `JointModelGroup`. Throughout MoveIt! the terms
  // "planning group" and "joint model group" are used interchangably.
  // static const std::string PLANNING_GROUP = "panda_arm";
  static const std::string PLANNING_GROUP = "arm";

  // The :move_group_interface:`MoveGroup` class can be easily
  // setup using just the name of the planning group you would like to control
  // and plan for.
  moveit::planning_interface::MoveGroupInterface move_group(PLANNING_GROUP);

  // We will use the :planning_scene_interface:`PlanningSceneInterface`
  // class to add and remove collision objects in our "virtual world" scene
  moveit::planning_interface::PlanningSceneInterface planning_scene_interface;

  // Raw pointers are frequently used to refer to the planning group for
  // improved performance.
  const robot_state::JointModelGroup *joint_model_group =
      move_group.getCurrentState()->getJointModelGroup(PLANNING_GROUP);

  // Setup the planner
  move_group.setPoseReferenceFrame("base_link");
  move_group.setPlannerId("RRTConnectkConfigDefault");
  move_group.setMaxVelocityScalingFactor(0.2);
  move_group.setMaxAccelerationScalingFactor(0.2);
  move_group.setNumPlanningAttempts(10);

  // Visualization
  // The package MoveItVisualTools provides many capabilties for visualizing
  // objects, robots, and trajectories in RViz as well as debugging tools such
  // as step-by-step introspection of a script

  //   namespace rvt = rviz_visual_tools;
  //   // moveit_visual_tools::MoveItVisualTools visual_tools("panda_link0");
  //   rviz_visual_tools::RvizVisualToolsPtr visual_tools_;
  //   visual_tools_.reset(new
  // rviz_visual_tools::RvizVisualTools("base_frame","/rviz_visual_markers"));
  //   visual_tools_.deleteAllMarkers();

  //   // Remote control is an introspection tool that allows users to step
  //   through a
  //   // high level script via buttons and keyboard shortcuts in RViz
  //   visual_tools.loadRemoteControl();

  //   // RViz provides many types of markers, in this demo we will use text,
  //   // cylinders, and spheres
  //   Eigen::Affine3d text_pose = Eigen::Affine3d::Identity();
  //   text_pose.translation().z() = 1.00;
  //   visual_tools.publishText(text_pose, "MoveGroupInterface Demo",
  //   rvt::WHITE,
  //                            rvt::XLARGE);

  //   // Batch publishing is used to reduce the number of messages being sent
  //   to
  //   // RViz for large visualizations
  //   visual_tools.trigger();

  // Getting Basic Information
  // ^^^^^^^^^^^^^^^^^^^^^^^^^
  //
  // We can print the name of the reference frame for this robot.
  ROS_INFO_NAMED("r7_test", "Reference frame: %s",
                 move_group.getPlanningFrame().c_str());

  // We can also print the name of the end-effector link for this group.
  ROS_INFO_NAMED("r7_test", "End effector link: %s",
                 move_group.getEndEffectorLink().c_str());

  // Start the demo
  // ^^^^^^^^^^^^^^^^^^^^^^^^^
  //   visual_tools.prompt(
  //       "Press 'next' in the RvizVisualToolsGui window to start the demo ");

  // Step 1) Planning to a Pose goal
  // ^^^^^^^^^^^^^^^^^^^^^^^
  // We can plan a motion for this group to a desired pose for the
  // end-effector.
  geometry_msgs::Pose target_pose1;
  target_pose1.orientation.x = 0.0;
  target_pose1.orientation.y = 1.0;
  target_pose1.orientation.z = 0.0;
  target_pose1.orientation.w = 0.001;
  target_pose1.position.x = 0.0;
  target_pose1.position.y = 0.4;
  target_pose1.position.z = 0.2;
  move_group.setPoseTarget(target_pose1);

  // Now, we call the planner to compute the plan and visualize it.
  // Note that we are just planning, not asking move_group
  // to actually move the robot.

  moveit::planning_interface::MoveGroupInterface::Plan my_plan;

  bool success =
      (move_group.plan(my_plan) == moveit_msgs::MoveItErrorCodes::SUCCESS);

  ROS_INFO_NAMED("r7_test", "Visualizing plan 1 (pose goal) %s",
                 success ? "" : "FAILED");

  //   // Visualizing plans
  //   // ^^^^^^^^^^^^^^^^^
  //   // We can also visualize the plan as a line with markers in RViz.
  //   ROS_INFO_NAMED("r7_test", "Visualizing plan 1 as trajectory line");
  //   visual_tools.publishAxisLabeled(target_pose1, "target_pose1");
  //   visual_tools.publishText(text_pose, "Pose Goal", rvt::WHITE,
  //   rvt::XLARGE); visual_tools.publishTrajectoryLine(my_plan.trajectory_,
  //   joint_model_group); visual_tools.trigger(); visual_tools.prompt("Press
  //   'next' in the RvizVisualToolsGui window to "
  //                       "execute the planned trajectory");

  // Moving to a pose goal
  // ^^^^^^^^^^^^^^^^^^^^^
  sleep(5);
  move_group.execute(my_plan);
  sleep(8);

  //   visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to
  //   plan "
  //                       "the Cartesian paths");

  // Step 2) Cartesian Paths
  // ^^^^^^^^^^^^^^^
  // You can plan a Cartesian path directly by specifying a list of waypoints
  // for the end-effector to go through. Note that we are starting
  // from the new start state above.  The initial pose (start state) does not
  // need to be added to the waypoint list but adding it can help with
  // visualizations
  geometry_msgs::Pose target_pose2 = move_group.getCurrentPose().pose;

  std::vector<geometry_msgs::Pose> waypoints;
  waypoints.push_back(target_pose2);

  target_pose2.position.x += 0.1;
  target_pose2.position.y -= 0.1;
  waypoints.push_back(target_pose2);

  target_pose2.position.y -= 0.1;
  target_pose2.position.x += 0.1;
  waypoints.push_back(target_pose2);

  target_pose2.position.z += 0.1;
  target_pose2.position.y += 0.1;
  target_pose2.position.x -= 0.1;
  waypoints.push_back(target_pose2);

  target_pose2.position.y += 0.1;
  target_pose2.position.x -= 0.1;
  waypoints.push_back(target_pose2);

  // Adding more points
  target_pose2.position.y -= 0.1;
  target_pose2.position.x += 0.1;
  waypoints.push_back(target_pose2);

  target_pose2.position.z -= 0.1;
  target_pose2.position.y -= 0.1;
  target_pose2.position.x += 0.1;
  waypoints.push_back(target_pose2);

  target_pose2.position.x += 0.1;
  waypoints.push_back(target_pose2);

  target_pose2.position.y += 0.1;
  target_pose2.position.x += 0.1;
  waypoints.push_back(target_pose2);

  target_pose2.position.x = 0.4;
  target_pose2.position.y = 0.0;
  target_pose2.position.z = 0.2;
  waypoints.push_back(target_pose2);

  //   target_pose2.position.x = 0.3;
  //   target_pose2.position.y = -0.3;
  //   target_pose2.position.z = 0.1;
  //   waypoints.push_back(target_pose2);

  // This beneath is the defined points that is NOT working properly
  //   geometry_msgs::Pose t1;
  //   geometry_msgs::Pose t2;
  //   geometry_msgs::Pose t3;
  //   geometry_msgs::Pose t4;

  //   t1.orientation.w = 0.001;
  //   t1.orientation.x = 0;
  //   t1.orientation.y = 1.0;
  //   t1.orientation.z = 0;
  //   t1.position.x = 0.0;
  //   t1.position.y = 0.4;
  //   t1.position.z = 0.2;

  //   t2.orientation.w = 0.001;
  //   t2.orientation.x = 0;
  //   t2.orientation.y = 1.0;
  //   t2.orientation.z = 0;
  //   t2.position.x = 0.3;
  //   t2.position.y = 0.3;
  //   t2.position.z = 0.2;

  //   t3.orientation.w = 0.001;
  //   t3.orientation.x = 0;
  //   t3.orientation.y = 1.0;
  //   t3.orientation.z = 0;
  //   t3.position.x = 0.4;
  //   t3.position.y = 0.0;
  //   t3.position.z = 0.2;

  //   t4.orientation.w = 0.001;
  //   t4.orientation.x = 0;
  //   t4.orientation.y = 1.0;
  //   t4.orientation.z = 0;
  //   t4.position.x = 0.3;
  //   t4.position.y = -0.3;
  //   t4.position.z = 0.2;

  //   waypoints.push_back(t1);
  //   waypoints.push_back(t2);
  //   waypoints.push_back(t3);
  //   waypoints.push_back(t4);

  for (geometry_msgs::Pose tp : waypoints) {
    ROS_INFO("!!! NB !!! Pose: %f, %f, %f", tp.position.x, tp.position.y,
             tp.position.z);
    ROS_INFO("!!! NB !!! Quarterion: %f, %f, %f, %f", tp.orientation.x,
             tp.orientation.y, tp.orientation.z, tp.orientation.w);
  }

  // Cartesian motions are frequently needed to be slower for actions such as
  // approach and retreat grasp motions. Here we demonstrate how to reduce the
  // speed of the robot arm via a scaling factor of the maxiumum speed of each
  // joint. Note this is not the speed of the end effector point.
  move_group.setMaxVelocityScalingFactor(0.5);

  // We want the Cartesian path to be interpolated at a resolution of 1 cm
  // which is why we will specify 0.01 as the max step in Cartesian
  // translation.  We will specify the jump threshold as 0.0, effectively
  // disabling it. Warning - disabling the jump threshold while operating real
  // hardware can cause large unpredictable motions of redundant joints and
  // could be a safety issue
  moveit_msgs::RobotTrajectory trajectory;
  const double jump_threshold = 0.0;
  const double eef_step = 0.01;
  double fraction = move_group.computeCartesianPath(waypoints, eef_step,
                                                    jump_threshold, trajectory);
  ROS_INFO_NAMED("r7_test",
                 "Visualizing plan 2 (Cartesian path) (%.2f%% acheived)",
                 fraction * 100.0);

  int count = 0;
  for (trajectory_msgs::JointTrajectoryPoint jtp :
       trajectory.joint_trajectory.points) {
    std::vector<double> accs =
        jtp.accelerations; // Comes from jtp.accelerations as float64[], as
                           // array
    std::vector<double> pos =
        jtp.positions; // Comes from jtp.positions as float64[], an array
    std::vector<double> eff =
        jtp.effort; // Comes from jtp.effort as float64[], an array
    std::vector<double> vel =
        jtp.velocities; // Comes from jtp.velocities as float64[], an array

    double secs = ros::Time::now().toSec();
    ros::Duration rd = jtp.time_from_start;
    secs = rd.toSec();
    ROS_INFO("Point nr. %i, Time: %f , Accs: %f %f %f %f %f %f, Pos: %f %f %f "
             "%f %f %f, Eff: %lu, Vel: %f %f %f %f %f %f",
             count, secs, accs[0], accs[1], accs[2], accs[3], accs[4], accs[5],
             pos[0], pos[1], pos[2], pos[3], pos[4], pos[5], eff.size(), vel[0],
             vel[1], vel[2], vel[3], vel[4], vel[5]);
    // ROS_INFO("Point nr. %i, Time: %f , Accs: %lu, Pos: %lu, Eff: %lu, Vel:
    // %lu",
    //          count, secs, accs.size(), pos.size(), eff.size(), vel.size());
    count += 1;
  }

  //   // Visualize the plan in RViz
  //   visual_tools.deleteAllMarkers();
  //   visual_tools.publishText(text_pose, "Cartesian path", rvt::WHITE,
  //                            rvt::XLARGE);
  //   visual_tools.publishPath(waypoints, rvt::LIME_GREEN, rvt::SMALL);
  //   for (std::size_t i = 0; i < waypoints.size(); ++i)
  //     visual_tools.publishAxisLabeled(waypoints[i], "pt" + std::to_string(i),
  //                                     rvt::LARGE);
  //   visual_tools.trigger();
  //   visual_tools.prompt(
  //       "Press 'next' in the RvizVisualToolsGui window to continue the
  //       demo");

  // Moving to a Constrained Goal
  // ^^^^^^^^^^^^^^^^^^^^^
  sleep(5);
  my_plan.trajectory_ = trajectory;
  if (fraction == 1.0) {
    move_group.execute(my_plan);
  } else
    ROS_WARN("Could not compute the cartesian path :( ");

  // END_TUTORIAL

  ros::shutdown();
  return 0;
}