// Following the MoveIT! Kinetic tutorial (depricated?)
// http://docs.ros.org/kinetic/api/moveit_tutorials/html/index.html

#include <moveit/move_group_interface/move_group_interface.h>
#include <moveit/planning_scene_interface/planning_scene_interface.h>

#include <moveit_msgs/DisplayRobotState.h>
#include <moveit_msgs/DisplayTrajectory.h>

#include <moveit_msgs/AttachedCollisionObject.h>
#include <moveit_msgs/CollisionObject.h>

#include <moveit_visual_tools/moveit_visual_tools.h>

// #### NB #### Legge til i panels i RViz den RVizVisualTools fyrst. 
// Må en fikse noke meir (?) Som for eksempel planning_scene_topic, plannigGroup eller trajectoryTopic? 

int main(int argc, char** argv)
{
    ros::init(argc, argv, "r7_test");
    ros::NodeHandle node_handle;
    ros::AsyncSpinner spinner(1);
    spinner.start();

    static const std::string PLANNING_GROUP = "arm";
    moveit::planning_interface::MoveGroupInterface move_group(PLANNING_GROUP);
    moveit::planning_interface::PlanningSceneInterface planning_scene_interface;
    const robot_state::JointModelGroup* joint_model_group = move_group.getCurrentState()->getJointModelGroup(PLANNING_GROUP);

    // From the Motion planning API Turotial (?) wtf.. This uses a "planner" of some kind to specifically tell RViz what is going on? Unnecessary?
    // http://docs.ros.org/kinetic/api/moveit_tutorials/html/doc/motion_planning_api/motion_planning_api_tutorial.html 


  // Visualization
  // The package MoveItVisualTools provides many capabilties for visualizing objects, robots,
  // and trajectories in RViz as well as debugging tools such as step-by-step introspection of a script
  //To use this, the RVizVisualToolsGui should be initiated in the RViz window adding it through the panels menu.
  namespace rvt = rviz_visual_tools;
  moveit_visual_tools::MoveItVisualTools visual_tools("base_link");
  visual_tools.deleteAllMarkers();

  // Remote control is an introspection tool that allows users to step through a high level script
  // via buttons and keyboard shortcuts in RViz
  visual_tools.loadRemoteControl();

  // RViz provides many types of markers, in this demo we will use text, cylinders, and spheres
  Eigen::Affine3d text_pose = Eigen::Affine3d::Identity();
  text_pose.translation().z() = 1.75;
  visual_tools.publishText(text_pose, "MoveGroupInterface Demo", rvt::WHITE, rvt::XLARGE);

  // Batch publishing is used to reduce the number of messages being sent to RViz for large visualizations
  visual_tools.trigger();

  // Getting Basic Information
  // We can print the name of the reference frame for this robot.
  ROS_INFO_NAMED("r7_test", "Reference frame: %s", move_group.getPlanningFrame().c_str());

  // We can also print the name of the end-effector link for this group.
  ROS_INFO_NAMED("r7_test", "End effector link: %s", move_group.getEndEffectorLink().c_str());

  // Start the demo
  // ^^^^^^^^^^^^^^^^^^^^^^^^^
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to start the demo");

  // Planning to a Pose goal
  geometry_msgs::Pose target_pose1;
  target_pose1.orientation.w = 0.001;
  target_pose1.orientation.y = 1.0;
  target_pose1.position.x = 0.3;
  target_pose1.position.y = 0.3;
  target_pose1.position.z = 0.2;
  move_group.setPoseTarget(target_pose1);

  // Now, we call the planner to compute the plan and visualize it.
  // Note that we are just planning, not asking move_group to actually move the robot.
  moveit::planning_interface::MoveGroupInterface::Plan my_plan;

  bool success = (move_group.plan(my_plan) == moveit::planning_interface::MoveItErrorCode::SUCCESS);  //moveit_msgs::MoveItErrorCodes::SUCCESS ??

  ROS_INFO_NAMED("r7_test", "Visualizing plan 1 (pose goal) %s", success ? "" : "FAILED");

  // Visualizing plans
  // We can also visualize the plan as a line with markers in RViz.
  ROS_INFO_NAMED("r7_test", "Visualizing plan 1 as trajectory line");
  visual_tools.publishAxisLabeled(target_pose1, "pose1");
  visual_tools.publishText(text_pose, "Pose Goal", rvt::WHITE, rvt::XLARGE);
  visual_tools.publishTrajectoryLine(my_plan.trajectory_, joint_model_group);
  visual_tools.trigger();
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to continue the demo");

  // Moving to a pose goal, actually moving the real robot
    sleep(5);
    move_group.move(); 

  // Planning to a joint-space goal
  // To start, we'll create an pointer that references the current robot's state.
  // RobotState is the object that contains all the current position/velocity/acceleration data.
  moveit::core::RobotStatePtr current_state = move_group.getCurrentState();

  // Next get the current set of joint values for the group.
  std::vector<double> joint_group_positions;
  current_state->copyJointGroupPositions(joint_model_group, joint_group_positions);

  // Now, let's modify one of the joints, plan to the new joint space goal and visualize the plan.
  joint_group_positions[0] = -1.57;  // radians
  move_group.setJointValueTarget(joint_group_positions);

  success = (move_group.plan(my_plan) == moveit::planning_interface::MoveItErrorCode::SUCCESS);
  ROS_INFO_NAMED("tutorial", "Visualizing plan 2 (joint space goal) %s", success ? "" : "FAILED");

  // Visualize the plan in RViz
  visual_tools.deleteAllMarkers();
  visual_tools.publishText(text_pose, "Joint Space Goal", rvt::WHITE, rvt::XLARGE);
  visual_tools.publishTrajectoryLine(my_plan.trajectory_, joint_model_group);
  visual_tools.trigger();
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to continue the demo");

//   // Planning with Path Constraints
//   // Path constraints can easily be specified for a link on the robot.
//   // Let's specify a path constraint and a pose goal for our group.
//   // First define the path constraint.
//   moveit_msgs::OrientationConstraint ocm;
//   ocm.link_name = "panda_link7";
//   ocm.header.frame_id = "panda_link0";
//   ocm.orientation.w = 1.0;
//   ocm.absolute_x_axis_tolerance = 0.1;
//   ocm.absolute_y_axis_tolerance = 0.1;
//   ocm.absolute_z_axis_tolerance = 0.1;
//   ocm.weight = 1.0;

//   // Now, set it as the path constraint for the group.
//   moveit_msgs::Constraints test_constraints;
//   test_constraints.orientation_constraints.push_back(ocm);
//   move_group.setPathConstraints(test_constraints);

//   // We will reuse the old goal that we had and plan to it.
//   // Note that this will only work if the current state already
//   // satisfies the path constraints. So, we need to set the start
//   // state to a new pose.
//   robot_state::RobotState start_state(*move_group.getCurrentState());
//   geometry_msgs::Pose start_pose2;
//   start_pose2.orientation.w = 1.0;
//   start_pose2.position.x = 0.55;
//   start_pose2.position.y = -0.05;
//   start_pose2.position.z = 0.8;
//   start_state.setFromIK(joint_model_group, start_pose2);
//   move_group.setStartState(start_state);

//   // Now we will plan to the earlier pose target from the new
//   // start state that we have just created.
//   move_group.setPoseTarget(target_pose1);

//   // Planning with constraints can be slow because every sample must call an inverse kinematics solver.
//   // Lets increase the planning time from the default 5 seconds to be sure the planner has enough time to succeed.
//   move_group.setPlanningTime(10.0);

//   success = (move_group.plan(my_plan) == moveit::planning_interface::MoveItErrorCode::SUCCESS);
//   ROS_INFO_NAMED("tutorial", "Visualizing plan 3 (constraints) %s", success ? "" : "FAILED");

//   // Visualize the plan in RViz
//   visual_tools.deleteAllMarkers();
//   visual_tools.publishAxisLabeled(start_pose2, "start");
//   visual_tools.publishAxisLabeled(target_pose1, "goal");
//   visual_tools.publishText(text_pose, "Constrained Goal", rvt::WHITE, rvt::XLARGE);
//   visual_tools.publishTrajectoryLine(my_plan.trajectory_, joint_model_group);
//   visual_tools.trigger();
//   visual_tools.prompt("next step");

//   // When done with the path constraint be sure to clear it.
//   move_group.clearPathConstraints();

// Since we set the start state we have to clear it before planning other paths
   move_group.setStartStateToCurrentState();

  // Cartesian Paths
  // ^^^^^^^^^^^^^^^
  // You can plan a Cartesian path directly by specifying a list of waypoints
  // for the end-effector to go through. Note that we are starting
  // from the new start state above.  The initial pose (start state) does not
  // need to be added to the waypoint list but adding it can help with visualizations
  geometry_msgs::PoseStamped target_pose3 = move_group.getCurrentPose().pose; // Or set this yourself ->PosStamped and then the pose function of this. 

  std::vector<geometry_msgs::Pose> waypoints;
  waypoints.push_back(target_pose3);

  target_pose3.position.z -= 0.2;
  waypoints.push_back(target_pose3);  // down

  target_pose3.position.y -= 0.2;
  waypoints.push_back(target_pose3);  // right

  target_pose3.position.z += 0.2;
  target_pose3.position.y += 0.1;
  target_pose3.position.x -= 0.1;
  waypoints.push_back(target_pose3);  // up and left

  move_group.setMaxVelocityScalingFactor(0.7);

  const moveit_msgs::RobotTrajectory trajectory;
  const double jump_threshold = 0.0;
  const double eef_step = 0.01;
  double fraction = move_group.computeCartesianPath(waypoints, eef_step, jump_threshold, trajectory);
  ROS_INFO_NAMED("r7_test", "Visualizing plan 4 (Cartesian path) (%.2f%% achieved)", fraction * 100.0);

  // Visualize the plan in RViz
  visual_tools.deleteAllMarkers();
  visual_tools.publishText(text_pose, "Joint Space Goal", rvt::WHITE, rvt::XLARGE);
  visual_tools.publishPath(waypoints, rvt::LIME_GREEN, rvt::SMALL);
  for (std::size_t i = 0; i < waypoints.size(); ++i)
    visual_tools.publishAxisLabeled(waypoints[i], "pt" + std::to_string(i), rvt::SMALL);
  visual_tools.trigger();
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to continue/finish the demo");
  sleep(10);

//   //Execution
//   move_group.execute(&trajectory); //Or my_plan ?? after my_plan.trajectory_ = trajectory
//   //my_plan.trajectory_ = trajectory;
//   //move_group.execute(my_plan);

//   //###################################### From google groups 2014 -> FOr defining the speed of the trajectory execution.
//   moveit_msgs::RobotTrajectory trajectory_msg;
//   group.setPlanningTime(10.0);
 
//   double fraction = group.computeCartesianPath(waypoints,
//                                                0.01,  // eef_step
//                                                0.0,   // jump_threshold
//                                                trajectory_msg, false);
//   // The trajectory needs to be modified so it will include velocities as well.
//   // First to create a RobotTrajectory object
//   robot_trajectory::RobotTrajectory rt(group.getCurrentState()->getRobotModel(), "arm_group");

//   // Second get a RobotTrajectory from trajectory
//   rt.setRobotTrajectoryMsg(*group.getCurrentState(), trajectory_msg);
 
//   // Thrid create a IterativeParabolicTimeParameterization object
//   trajectory_processing::IterativeParabolicTimeParameterization iptp;

//   // Fourth compute computeTimeStamps
//   success = iptp.computeTimeStamps(rt);
//   ROS_INFO("Computed time stamp %s",success?"SUCCEDED":"FAILED");

//   // Get RobotTrajectory_msg from RobotTrajectory
//   rt.getRobotTrajectoryMsg(trajectory_msg);

//   // Finally plan and execute the trajectory
//   plan.trajectory_ = trajectory_msg;
//   ROS_INFO("Visualizing plan 4 (cartesian path) (%.2f%% acheived)",fraction * 100.0);   
//   sleep(5.0);
//   group.execute(plan);
// ###########################################



//   // Adding/Removing Objects and Attaching/Detaching Objects
//   // Define a collision object ROS message.
//   moveit_msgs::CollisionObject collision_object;
//   collision_object.header.frame_id = move_group.getPlanningFrame();

//   // The id of the object is used to identify it.
//   collision_object.id = "box1";

//   // Define a box to add to the world.
//   shape_msgs::SolidPrimitive primitive;
//   primitive.type = primitive.BOX;
//   primitive.dimensions.resize(3);
//   primitive.dimensions[0] = 0.4;
//   primitive.dimensions[1] = 0.1;
//   primitive.dimensions[2] = 0.4;

//   // Define a pose for the box (specified relative to frame_id)
//   geometry_msgs::Pose box_pose;
//   box_pose.orientation.w = 1.0;
//   box_pose.position.x = 0.4;
//   box_pose.position.y = -0.2;
//   box_pose.position.z = 1.0;

//   collision_object.primitives.push_back(primitive);
//   collision_object.primitive_poses.push_back(box_pose);
//   collision_object.operation = collision_object.ADD;

//   std::vector<moveit_msgs::CollisionObject> collision_objects;
//   collision_objects.push_back(collision_object);

//   // Now, let's add the collision object into the world
//   ROS_INFO_NAMED("tutorial", "Add an object into the world");
//   planning_scene_interface.addCollisionObjects(collision_objects);

//   // Show text in RViz of status
//   visual_tools.publishText(text_pose, "Add object", rvt::WHITE, rvt::XLARGE);
//   visual_tools.trigger();

//   // Wait for MoveGroup to recieve and process the collision object message
//   visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to once the collision object appears in RViz");

//   // Now when we plan a trajectory it will avoid the obstacle
//   move_group.setStartState(*move_group.getCurrentState());
//   geometry_msgs::Pose another_pose;
//   another_pose.orientation.w = 1.0;
//   another_pose.position.x = 0.4;
//   another_pose.position.y = -0.4;
//   another_pose.position.z = 0.9;
//   move_group.setPoseTarget(another_pose);

//   success = (move_group.plan(my_plan) == moveit::planning_interface::MoveItErrorCode::SUCCESS);
//   ROS_INFO_NAMED("tutorial", "Visualizing plan 5 (pose goal move around cuboid) %s", success ? "" : "FAILED");

//   // Visualize the plan in RViz
//   visual_tools.deleteAllMarkers();
//   visual_tools.publishText(text_pose, "Obstacle Goal", rvt::WHITE, rvt::XLARGE);
//   visual_tools.publishTrajectoryLine(my_plan.trajectory_, joint_model_group);
//   visual_tools.trigger();
//   visual_tools.prompt("next step");

//   // Now, let's attach the collision object to the robot.
//   ROS_INFO_NAMED("tutorial", "Attach the object to the robot");
//   move_group.attachObject(collision_object.id);

//   // Show text in RViz of status
//   visual_tools.publishText(text_pose, "Object attached to robot", rvt::WHITE, rvt::XLARGE);
//   visual_tools.trigger();

//   /* Wait for MoveGroup to recieve and process the attached collision object message */
//   visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to once the collision object attaches to the "
//                       "robot");

//   // Now, let's detach the collision object from the robot.
//   ROS_INFO_NAMED("tutorial", "Detach the object from the robot");
//   move_group.detachObject(collision_object.id);

//   // Show text in RViz of status
//   visual_tools.publishText(text_pose, "Object dettached from robot", rvt::WHITE, rvt::XLARGE);
//   visual_tools.trigger();

//   /* Wait for MoveGroup to recieve and process the attached collision object message */
//   visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to once the collision object detaches to the "
//                       "robot");

//   // Now, let's remove the collision object from the world.
//   ROS_INFO_NAMED("tutorial", "Remove the object from the world");
//   std::vector<std::string> object_ids;
//   object_ids.push_back(collision_object.id);
//   planning_scene_interface.removeCollisionObjects(object_ids);

//   // Show text in RViz of status
//   visual_tools.publishText(text_pose, "Object removed", rvt::WHITE, rvt::XLARGE);
//   visual_tools.trigger();

//   /* Wait for MoveGroup to recieve and process the attached collision object message */
//   visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to once the collision object disapears");

  ros::shutdown();
  return 0;
}